import json

from numbers import Number
from logging import LogRecord
from dataclasses import dataclass, field, fields
from typing import Any, Dict, Optional


@dataclass
class AlertMessage:
    name: str
    msg: str
    levelno: int = 30
    levelname: str = 'WARNING'
    alert_type: str = 'logger'
    pathname: Optional[str] = None
    filename: Optional[str] = None
    module: Optional[str] = None
    lineno: Optional[int] = None
    funcName: Optional[str] = None
    created: Optional[float] = None
    processName: Optional[str] = None
    extra: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_dict(cls, _record_dict: Dict[str, Any]) -> 'AlertMessage':
        # Create a AlertMessage instance from a dict with the matching fields
        record_dict = {
            k: v
            for k, v in _record_dict.items()
            if v is not None and k in FIELD_NAMES
        }
        log_message = cls(**record_dict)

        # Add extra data to the extra field
        # data that is not part of the AlertMessage dataclass fields
        extra_data = {
            k: v
            for k, v in _record_dict.items()
            if v is not None and k not in FIELD_NAMES
        }
        log_message.extra = extra_data

        return log_message

    @classmethod
    def log_record_to_dict(cls, log_record: LogRecord) -> Dict[str, Any]:

        def jsonify_value(value: Any) -> Any:
            if isinstance(value, (str, Number)):
                return value
            return json.dumps(value, separators=(',', ':'), default=str)

        record_dict = {
            k: jsonify_value(v)
            for k, v in log_record.__dict__.items()
            if (k not in LOG_RECORD_EXC_FIELDS and not k.startswith('_')
                and v is not None)
        }

        return record_dict


# Process constants only one time
FIELD_NAMES = {f.name for f in fields(AlertMessage)}
LOG_RECORD_EXC_FIELDS = {
    'args', 'msecs', 'process', 'relativeCreated', 'thread', 'threadName',
    'exc_info'
}



import os
import sys
import json
import inspect
import logging
import logging.handlers
import traceback

from pathlib import Path
from dataclasses import fields
from arb_logger.alert_message import AlertMessage

import coloredlogs

from redis import Redis


def get_redis_log_client() -> Redis:
    # Will be replaced by another Redis client prolly
    return Redis(host='localhost', port=6379, db=0, decode_responses=True)


def get_redis_log_key(name: str):
    return f'logs:{name}'


class RedisHandler(logging.Handler):

    def __init__(self, redis_client, redis_key):
        super().__init__()
        self.redis_client: Redis = redis_client
        self.redis_key = redis_key

        self.alert_fields = {f.name for f in fields(AlertMessage)}

    def emit(self, record: logging.LogRecord):
        try:
            record_dict = AlertMessage.log_record_to_dict(record)
            #! split : to get the stream name being 'logs' for grafana
            self.redis_client.xadd(self.redis_key.split(':')[0], record_dict)
            self.redis_client.publish(self.redis_key,
                                      json.dumps(record_dict,
                                                 separators=(',', ':'),
                                                 default=str))  # Add this line
        except Exception:
            self.handleError(record)


STDOUT_LEVEL_NUM = 25
logging.addLevelName(STDOUT_LEVEL_NUM, 'STDOUT')


def stdout(self, message, *args, **kwargs):
    if self.isEnabledFor(STDOUT_LEVEL_NUM):
        self._log(STDOUT_LEVEL_NUM, message, args, **kwargs)


logging.Logger.stdout = stdout


class LoggerStdout:

    def __init__(self, logger):
        self.logger = logger
        self.stdout = sys.stdout

    def write(self, message):
        if message.rstrip() != '':
            self.logger.stdout('\n' + message.rstrip())

    def flush(self):
        self.stdout.flush()

    def isatty(self):
        return self.stdout.isatty()


def get_logger_name():
    # Get the name of the calling module
    name = inspect.getmodule(inspect.currentframe().f_back).__name__
    if name in ('__main__', '__init__'):
        raise ValueError(
            'Please provide an explicit name for the logger when called from __main__ or __init__.'
        )

    return name


def get_logger(name: str = None,
               level: int = logging.DEBUG,
               path: Path = None,
               log_in_file: bool = True,
               short: bool = False,
               redis_handler: bool = True,
               custom_redis_client: Redis = None):

    logger = logging.getLogger(name or get_logger_name())

    if not logger.hasHandlers():
        logger.setLevel(level)
        logger.propagate = False

        # Set up the custom formatter
        fmt = '%(asctime)s [%(levelname)s] %(name)s - %(message)s - (%(filename)s:%(lineno)d)'
        formatter = logging.Formatter(fmt)

        # Set up coloredlogs with the custom formatter
        coloredlogs.install(level=level, logger=logger, fmt=fmt)

        # Set up RedisHandler
        if redis_handler:
            redis_client = custom_redis_client or get_redis_log_client()
            redis_key = get_redis_log_key(name)
            redis_handler = RedisHandler(redis_client, redis_key)
            redis_handler.setLevel(logging.WARNING)
            redis_handler.setFormatter(formatter)
            logger.addHandler(redis_handler)

        # Set up FileHandler
        if log_in_file:
            log_dir = path or Path(os.environ.get('ARB_LOGS_PATH', 'logs'))
            log_dir.mkdir(parents=True, exist_ok=True)

            # Get the rootname of the calling module/logger name
            # use it as the log filename to regroup multiple sublogger
            # under the same file
            _filename = name.split('.')[0]
            if short:
                filename = f'{_filename}.log'
            else:
                pid = os.getpid()
                filename = f'{_filename}.{pid}.log'

            file_handler = logging.handlers.TimedRotatingFileHandler(
                log_dir / filename,
                when='midnight',
                interval=1,
                backupCount=3,
                utc=True)
            file_handler.setFormatter(formatter)
            logger.addHandler(file_handler)

        # Log uncaught exceptions using sys.excepthook
        def log_uncaught_exceptions(exc_type, exc_value, exc_traceback):
            if exc_type is KeyboardInterrupt:
                sys.__excepthook__(exc_type, exc_value, exc_traceback)
                return
            logger.critical(''.join(
                traceback.format_exception(exc_type, exc_value,
                                           exc_traceback)))

        sys.stdout = LoggerStdout(logger)
        sys.excepthook = log_uncaught_exceptions

    return logger



from setuptools import setup, find_packages

setup(
    name='arb_logger',
    author='arb',
    author_email='arb.trops@gmail.com',
    description="A custom logger with Redis integration",
    url="https://github.com/Acid3croco/arb-trops-services",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    version='2.1.3',
    packages=find_packages(),
    install_requires=['coloredlogs', 'redis', 'pync'],
    entry_points={
        'console_scripts': [
            'arb_alerts = arb_logger.arb_alerts:main',
        ]
    },
)



import json
import logging
import platform

from arb_logger.alert_message import AlertMessage
from arb_logger.logger import get_logger, get_redis_log_client, get_redis_log_key

if platform.system() == 'Darwin':
    import pync

LOGGER = get_logger('arb_alerts', short=True, redis_handler=False)


class ArbAlerts:

    def __init__(self):
        self.redis_client = get_redis_log_client()

    def listen(self):
        self.pubsub = self.redis_client.pubsub(ignore_subscribe_messages=True)
        pattern = get_redis_log_key('*')
        self.pubsub.psubscribe(pattern)

        for message in self.pubsub.listen():
            if message['type'] == 'pmessage':
                if message['channel'] == 'logs:arb_alerts':
                    # Prevent recursive logging of arb_alerts's messages
                    continue
                try:
                    record_dict = json.loads(message['data'])
                    log_message = AlertMessage.from_dict(record_dict)
                    self.log_alert_message(log_message)
                except Exception as e:
                    LOGGER.exception(
                        f"Error while processing log message: {e}")

    def log_alert_message(self, alert_message: AlertMessage):
        # Convert back AlertMessage to LogRecord to be handled by local logger easily
        msg = alert_message.msg
        if alert_message.extra:
            msg += f" {str(alert_message.extra)}"
        log_record = logging.LogRecord(name=alert_message.name,
                                       level=alert_message.levelno,
                                       pathname=alert_message.pathname,
                                       lineno=alert_message.lineno,
                                       msg=msg,
                                       func=alert_message.funcName,
                                       args=None,
                                       exc_info=None)
        LOGGER.handle(log_record)

        if alert_message.levelno >= logging.ERROR:
            """
            make this async
            """
            self.send_macos_notification(alert_message)

    def send_macos_notification(self, log_message: AlertMessage):
        if platform.system() == 'Darwin':
            title = f"{log_message.levelname} - {log_message.name}"
            subtitle = f"{log_message.filename}:{log_message.lineno}"
            message = log_message.msg
            pync.notify(message, title=title, subtitle=subtitle, wait=True)


def main():
    LOGGER.info('Starting arb_alerts')
    arb_alerts = ArbAlerts()
    LOGGER.info('Listening for log messages...')
    arb_alerts.listen()


if __name__ == '__main__':
    main()
